import warnings
from copy import copy
from typing import TYPE_CHECKING, Optional, Sequence

import aesara
import aesara.tensor as at
import numpy as np
from aesara.compile.builders import OpFromGraph
from aesara.graph.basic import Apply, Constant
from aesara.graph.op import Op
from aesara.tensor.basic import make_vector
from aesara.tensor.random.utils import broadcast_params, normalize_size_param
from aesara.tensor.var import TensorVariable

from aeppl.abstract import MeasurableVariable, _get_measurable_outputs

if TYPE_CHECKING:
    from aesara.tensor.random.utils import RandomStream


class DiracDelta(Op):
    """An `Op` that represents a Dirac-delta distribution."""

    __props__ = ("rtol", "atol")

    def __init__(self, rtol=1e-5, atol=1e-8):
        self.rtol = rtol
        self.atol = atol

    def make_node(self, x):
        x = at.as_tensor(x)
        return Apply(self, [x], [x.type()])

    def do_constant_folding(self, fgraph, node):
        # Without this, the `Op` would be removed from the graph during
        # canonicalization
        return False

    def perform(self, node, inp, out):
        (x,) = inp
        (z,) = out
        warnings.warn(
            "DiracDelta is a dummy Op that shouldn't be used in a compiled graph"
        )
        z[0] = x

    def infer_shape(self, fgraph, node, input_shapes):
        return input_shapes


dirac_delta = DiracDelta()

MeasurableVariable.register(DiracDelta)


def non_constant(x):
    x = at.as_tensor_variable(x)
    if isinstance(x, Constant):
        # XXX: This isn't good for `size` parameters, because it could result
        # in `at.get_vector_length` exceptions.
        res = x.type()
        res.tag = copy(res.tag)
        if aesara.config.compute_test_value != "off":
            res.tag.test_value = x.data
        res.name = x.name
        return res
    else:
        return x


def switching_process(
    comp_rvs: Sequence[TensorVariable],
    states: TensorVariable,
):
    """Construct a switching process over arbitrary univariate mixtures and a state sequence.

    This simply constructs a graph of the following form:

        at.stack(comp_rvs)[states, *idx]

    where ``idx`` makes sure that `states` selects mixture components along all
    the other axes.

    Parameters
    ----------
    comp_rvs
        A list containing `MeasurableVariable` objects for each mixture component.
    states
        The hidden state sequence.  It should have a number of states
        equal to the size of `comp_dists`.

    """

    states = at.as_tensor(states, dtype=np.int64)
    comp_rvs_bcast = at.broadcast_arrays(*[at.as_tensor(rv) for rv in comp_rvs])
    M_rv = at.stack(comp_rvs_bcast)
    indices = (states,) + tuple(at.arange(d) for d in tuple(M_rv.shape)[1:])
    rv_var = M_rv[indices]
    return rv_var


class DiscreteMarkovChainFactory(OpFromGraph):
    """An `Op` constructed from an Aesara graph that represents a discrete Markov chain.

    This "composite" `Op` allows us to mark a sub-graph as measurable and
    assign a `_logprob` dispatch implementation.

    As far as broadcasting is concerned, this `Op` has the following
    `RandomVariable`-like properties:

        ndim_supp = 1
        ndims_params = (3, 1)

    TODO: It would be nice to express this as a `Blockwise` `Op`.
    """

    default_output = 0


MeasurableVariable.register(DiscreteMarkovChainFactory)


@_get_measurable_outputs.register(DiscreteMarkovChainFactory)
def _get_measurable_outputs_DiscreteMarkovChainFactory(op, node):
    return [node.outputs[0]]


def create_discrete_mc_op(srng, size, Gammas, gamma_0):
    """Construct a `DiscreteMarkovChainFactory` `Op`.

    This returns a `Scan` that performs the follow:

        states[0] = categorical(gamma_0)
        for t in range(1, N):
            states[t] = categorical(Gammas[t, state[t-1]])

    The Aesara graph representing the above is wrapped in an `OpFromGraph` so
    that we can easily assign it a specific log-probability.

    TODO: Eventually, AePPL should be capable of parsing more sophisticated
    `Scan`s and producing nearly the same log-likelihoods, and the use of
    `OpFromGraph` will no longer be necessary.

    """

    # Again, we need to preserve the length of this symbolic vector, so we do
    # this.
    size_param = make_vector(
        *[non_constant(size[i]) for i in range(at.get_vector_length(size))]
    )
    size_param.name = "size"

    # We make shallow copies so that unwanted ancestors don't appear in the
    # graph.
    Gammas_param = non_constant(Gammas).clone()
    Gammas_param.name = "Gammas_param"

    gamma_0_param = non_constant(gamma_0).clone()
    gamma_0_param.name = "gamma_0_param"

    bcast_Gammas_param, bcast_gamma_0_param = broadcast_params(
        (Gammas_param, gamma_0_param), (3, 1)
    )

    # Sample state 0 in each state sequence
    state_0 = srng.categorical(
        bcast_gamma_0_param,
        size=tuple(size_param) + tuple(bcast_gamma_0_param.shape[:-1]),
        # size=at.join(0, size_param, bcast_gamma_0_param.shape[:-1]),
    )

    N = bcast_Gammas_param.shape[-3]
    states_shape = tuple(state_0.shape) + (N,)

    bcast_Gammas_param = at.broadcast_to(
        bcast_Gammas_param, states_shape + tuple(bcast_Gammas_param.shape[-2:])
    )

    def loop_fn(n, state_nm1, Gammas_inner):
        gamma_t = Gammas_inner[..., n, :, :]
        idx = tuple(at.ogrid[[slice(None, d) for d in tuple(state_0.shape)]]) + (
            state_nm1.T,
        )
        gamma_t = gamma_t[idx]
        state_n = srng.categorical(gamma_t)
        return state_n.T

    res, updates = aesara.scan(
        loop_fn,
        outputs_info=[{"initial": state_0.T, "taps": [-1]}],
        sequences=[at.arange(N)],
        non_sequences=[bcast_Gammas_param],
        # strict=True,
    )

    # TODO FIXME: This is unreliable and needs to be replaced with explicit
    # update support in `OpFromGraph`.
    update_outputs = [state_0.owner.inputs[0].default_update]
    update_outputs.extend(updates.values())

    return (
        DiscreteMarkovChainFactory(
            [size_param, Gammas_param, gamma_0_param],
            [res.T] + update_outputs,
            inline=True,
            on_unused_input="ignore",
        ),
        updates,
    )


def discrete_markov_chain(
    Gammas: TensorVariable,
    gamma_0: TensorVariable,
    size=None,
    srng: Optional["RandomStream"] = None,
    **kwargs
):
    """Construct a first-order discrete Markov chain distribution.

    This defines a random vector that consists of state indicator values (i.e.
    ``0`` to ``M - 1`` where ``M`` is the dimensionality of the state space)
    that are driven by a discrete Markov chain.

    Given an array of transition probability matrices ``Gamma`` and initial
    state probabilities ``gamma0``, `discrete_markov_chain` represents the
    probability distribution of ``states`` defined by:

    .. code::

        states[0] = categorical(gamma_0)
        for t in range(1, N):
            states[t] = categorical(Gammas[t, state[t-1]])

    Example
    -------

    .. code::

        import aesara.tensor as at
        import numpy as np

        num_steps = 10
        Gammas_base = np.array([[.5, .5], [.5, .5]])
        Gammas = np.broadcast(Gammas_base, (num_steps, 2, 2))
        gamma0 = np.r_[0.5, 0.5]

        dmc, updates = discrete_markov_chain(test_Gamma, test_gamma_0)


    Parameters
    ----------
    Gammas
        An array of transition probability matrices.  `Gammas` takes the
        shape ``... x N x M x M`` for a state sequence of length ``N`` having
        ``M``-many distinct states.  Each row, ``r``, in a transition probability
        matrix gives the probability of transitioning from state ``r`` to each
        other state.
    gamma_0
        The initial state probabilities.  The last dimension should be length ``M``,
        i.e. the number of distinct states.
    """
    gamma_0 = at.as_tensor_variable(gamma_0)

    assert Gammas.ndim >= 3

    Gammas = at.as_tensor_variable(Gammas)

    size = normalize_size_param(size)

    if srng is None:
        srng = at.random.RandomStream()

    dmc_op, updates = create_discrete_mc_op(srng, size, Gammas, gamma_0)
    rv_var = dmc_op(size, Gammas, gamma_0)

    # TODO FIXME: This is unreliable and needs to be replaced with explicit
    # update support in `OpFromGraph`.
    updates = {
        rv_var.owner.inputs[-2]: rv_var.owner.outputs[-2],
        rv_var.owner.inputs[-1]: rv_var.owner.outputs[-1],
    }

    testval = kwargs.pop("testval", None)

    if testval is not None:
        rv_var.tag.test_value = testval

    return rv_var, updates



import aesara
import aesara.tensor as at
import numpy as np
import pytest
import scipy.stats.distributions as sp
from aesara.graph.basic import Variable, equal_computations
from aesara.tensor.random.basic import CategoricalRV
from aesara.tensor.shape import shape_tuple
from aesara.tensor.subtensor import as_index_constant

from aeppl.dists import dirac_delta
from aeppl.joint_logprob import conditional_logprob, joint_logprob
from aeppl.mixture import MixtureRV, expand_indices
from aeppl.rewriting import construct_ir_fgraph
from tests.test_logprob import scipy_logprob
from tests.utils import assert_no_rvs


def test_mixture_basics():
    srng = at.random.RandomStream(29833)

    def create_mix_model(size, axis):
        X_rv = srng.normal(0, 1, size=size, name="X")
        Y_rv = srng.gamma(0.5, 0.5, size=size, name="Y")

        p_at = at.scalar("p")
        p_at.tag.test_value = 0.5

        I_rv = srng.bernoulli(p_at, size=size, name="I")

        if isinstance(axis, Variable):
            M_rv = at.join(axis, X_rv, Y_rv)[I_rv]
        else:
            M_rv = at.stack([X_rv, Y_rv], axis=axis)[I_rv]

        M_rv.name = "M"

        return locals()

    axis_at = at.lscalar("axis")
    axis_at.tag.test_value = 0
    env = create_mix_model((2,), axis_at)
    I_rv = env["I_rv"]
    M_rv = env["M_rv"]

    with pytest.raises(NotImplementedError):
        joint_logprob(M_rv, I_rv)


@aesara.config.change_flags(compute_test_value="warn")
@pytest.mark.parametrize(
    "op_constructor",
    [
        lambda _I, _X, _Y: at.stack([_X, _Y])[_I],
        lambda _I, _X, _Y: at.switch(_I, _X, _Y),
    ],
)
def test_compute_test_value(op_constructor):
    srng = at.random.RandomStream(29833)

    X_rv = srng.normal(0, 1, name="X")
    Y_rv = srng.gamma(0.5, 0.5, name="Y")

    p_at = at.scalar("p")
    p_at.tag.test_value = 0.3

    I_rv = srng.bernoulli(p_at, name="I")

    M_rv = op_constructor(I_rv, X_rv, Y_rv)
    M_rv.name = "M"

    M_logps, _ = conditional_logprob(M_rv, I_rv)

    del M_rv.tag.test_value

    M_logp = at.add(*M_logps.values())

    assert isinstance(M_logp.tag.test_value, np.ndarray)


@pytest.mark.parametrize(
    "p_val, size",
    [
        (np.array(0.0, dtype=aesara.config.floatX), ()),
        (np.array(1.0, dtype=aesara.config.floatX), ()),
        (np.array(0.0, dtype=aesara.config.floatX), (2,)),
        (np.array(1.0, dtype=aesara.config.floatX), (2, 1)),
        (np.array(1.0, dtype=aesara.config.floatX), (2, 3)),
        (np.array([0.1, 0.9], dtype=aesara.config.floatX), (2, 3)),
    ],
)
def test_hetero_mixture_binomial(p_val, size):
    srng = at.random.RandomStream(29833)

    X_rv = srng.normal(0, 1, size=size, name="X")
    Y_rv = srng.gamma(0.5, 0.5, size=size, name="Y")

    if np.ndim(p_val) == 0:
        p_at = at.scalar("p")
        p_at.tag.test_value = p_val
        I_rv = srng.bernoulli(p_at, size=size, name="I")
        p_val_1 = p_val
    else:
        p_at = at.vector("p")
        p_at.tag.test_value = np.array(p_val, dtype=aesara.config.floatX)
        I_rv = srng.categorical(p_at, size=size, name="I")
        p_val_1 = p_val[1]

    M_rv = at.stack([X_rv, Y_rv])[I_rv]
    M_rv.name = "M"

    M_logps, (M_vv, I_vv) = conditional_logprob(M_rv, I_rv)
    M_logp = at.add(*M_logps.values())

    M_logp_fn = aesara.function([p_at, M_vv, I_vv], M_logp)

    assert_no_rvs(M_logp_fn.maker.fgraph.outputs[0])

    decimals = 6 if aesara.config.floatX == "float64" else 4

    test_val_rng = np.random.RandomState(3238)

    bern_sp = sp.bernoulli(p_val_1)
    norm_sp = sp.norm(loc=0, scale=1)
    gamma_sp = sp.gamma(0.5, scale=1.0 / 0.5)

    for i in range(10):
        i_val = bern_sp.rvs(size=size, random_state=test_val_rng)
        x_val = norm_sp.rvs(size=size, random_state=test_val_rng)
        y_val = gamma_sp.rvs(size=size, random_state=test_val_rng)

        component_logps = np.stack([norm_sp.logpdf(x_val), gamma_sp.logpdf(y_val)])[
            i_val
        ]
        exp_obs_logps = component_logps + bern_sp.logpmf(i_val)

        m_val = np.stack([x_val, y_val])[i_val]
        logp_vals = M_logp_fn(p_val, m_val, i_val)

        np.testing.assert_almost_equal(logp_vals, exp_obs_logps, decimal=decimals)


@pytest.mark.parametrize(
    "X_args, Y_args, Z_args, p_val, comp_size, idx_size, extra_indices, join_axis",
    [
        # Scalar mixture components, scalar index
        (
            (
                np.array(0, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            (
                np.array(0.5, dtype=aesara.config.floatX),
                np.array(0.5, dtype=aesara.config.floatX),
            ),
            (
                np.array(100, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            np.array([0.1, 0.5, 0.4], dtype=aesara.config.floatX),
            (),
            (),
            (),
            0,
        ),
        # Scalar mixture components, vector index
        (
            (
                np.array(0, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            (
                np.array(0.5, dtype=aesara.config.floatX),
                np.array(0.5, dtype=aesara.config.floatX),
            ),
            (
                np.array(100, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            np.array([0.1, 0.5, 0.4], dtype=aesara.config.floatX),
            (),
            (6,),
            (),
            0,
        ),
        (
            (
                np.array([0, -100], dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            (
                np.array([0.5, 1], dtype=aesara.config.floatX),
                np.array([0.5, 1], dtype=aesara.config.floatX),
            ),
            (
                np.array([100, 1000], dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            np.array([[0.1, 0.5, 0.4], [0.4, 0.1, 0.5]], dtype=aesara.config.floatX),
            (2,),
            (2,),
            (),
            0,
        ),
        (
            (
                np.array([0, -100], dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            (
                np.array([0.5, 1], dtype=aesara.config.floatX),
                np.array([0.5, 1], dtype=aesara.config.floatX),
            ),
            (
                np.array([100, 1000], dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            np.array([[0.1, 0.5, 0.4], [0.4, 0.1, 0.5]], dtype=aesara.config.floatX),
            None,
            None,
            (),
            0,
        ),
        (
            (
                np.array(0, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            (
                np.array(0.5, dtype=aesara.config.floatX),
                np.array(0.5, dtype=aesara.config.floatX),
            ),
            (
                np.array(100, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            np.array([0.1, 0.5, 0.4], dtype=aesara.config.floatX),
            (),
            (),
            (),
            0,
        ),
        (
            (
                np.array(0, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            (
                np.array(0.5, dtype=aesara.config.floatX),
                np.array(0.5, dtype=aesara.config.floatX),
            ),
            (
                np.array(100, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            np.array([0.1, 0.5, 0.4], dtype=aesara.config.floatX),
            (2,),
            (2,),
            (),
            0,
        ),
        (
            (
                np.array(0, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            (
                np.array(0.5, dtype=aesara.config.floatX),
                np.array(0.5, dtype=aesara.config.floatX),
            ),
            (
                np.array(100, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            np.array([0.1, 0.5, 0.4], dtype=aesara.config.floatX),
            (2, 3),
            (2, 3),
            (),
            0,
        ),
        (
            (
                np.array(0, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            (
                np.array(0.5, dtype=aesara.config.floatX),
                np.array(0.5, dtype=aesara.config.floatX),
            ),
            (
                np.array(100, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            np.array([0.1, 0.5, 0.4], dtype=aesara.config.floatX),
            (2, 3),
            (),
            (),
            0,
        ),
        (
            (
                np.array(0, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            (
                np.array(0.5, dtype=aesara.config.floatX),
                np.array(0.5, dtype=aesara.config.floatX),
            ),
            (
                np.array(100, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            np.array([0.1, 0.5, 0.4], dtype=aesara.config.floatX),
            (3,),
            (3,),
            (slice(None),),
            1,
        ),
        (
            (
                np.array(0, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            (
                np.array(0.5, dtype=aesara.config.floatX),
                np.array(0.5, dtype=aesara.config.floatX),
            ),
            (
                np.array(100, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            np.array([0.1, 0.5, 0.4], dtype=aesara.config.floatX),
            (5,),
            (5,),
            (np.arange(5),),
            0,
        ),
        (
            (
                np.array(0, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            (
                np.array(0.5, dtype=aesara.config.floatX),
                np.array(0.5, dtype=aesara.config.floatX),
            ),
            (
                np.array(100, dtype=aesara.config.floatX),
                np.array(1, dtype=aesara.config.floatX),
            ),
            np.array([0.1, 0.5, 0.4], dtype=aesara.config.floatX),
            (5,),
            (5,),
            (np.arange(5), None),
            0,
        ),
    ],
)
def test_hetero_mixture_categorical(
    X_args, Y_args, Z_args, p_val, comp_size, idx_size, extra_indices, join_axis
):
    srng = at.random.RandomStream(29833)

    X_rv = srng.normal(*X_args, size=comp_size, name="X")
    Y_rv = srng.gamma(*Y_args, size=comp_size, name="Y")
    Z_rv = srng.normal(*Z_args, size=comp_size, name="Z")

    p_at = at.as_tensor(p_val).type()
    p_at.name = "p"
    p_at.tag.test_value = np.array(p_val, dtype=aesara.config.floatX)
    I_rv = srng.categorical(p_at, size=idx_size, name="I")

    indices_at = list(extra_indices)
    indices_at.insert(join_axis, I_rv)
    indices_at = tuple(indices_at)

    M_rv = at.stack([X_rv, Y_rv, Z_rv], axis=join_axis)[indices_at]
    M_rv.name = "M"

    logp_parts, (M_vv, I_vv) = conditional_logprob(M_rv, I_rv)

    I_logp_fn = aesara.function([p_at, I_vv], logp_parts[I_rv])
    M_logp_fn = aesara.function([M_vv, I_vv], logp_parts[M_rv])

    assert_no_rvs(I_logp_fn.maker.fgraph.outputs[0])
    assert_no_rvs(M_logp_fn.maker.fgraph.outputs[0])

    decimals = 6 if aesara.config.floatX == "float64" else 4

    test_val_rng = np.random.RandomState(3238)

    norm_1_sp = sp.norm(loc=X_args[0], scale=X_args[1])
    gamma_sp = sp.gamma(Y_args[0], scale=1 / Y_args[1])
    norm_2_sp = sp.norm(loc=Z_args[0], scale=Z_args[1])

    for i in range(10):
        i_val = CategoricalRV.rng_fn(test_val_rng, p_val, idx_size)

        indices_val = list(extra_indices)
        indices_val.insert(join_axis, i_val)
        indices_val = tuple(indices_val)

        x_val = norm_1_sp.rvs(size=comp_size, random_state=test_val_rng)
        y_val = gamma_sp.rvs(size=comp_size, random_state=test_val_rng)
        z_val = norm_2_sp.rvs(size=comp_size, random_state=test_val_rng)

        component_logps = np.stack(
            [norm_1_sp.logpdf(x_val), gamma_sp.logpdf(y_val), norm_2_sp.logpdf(z_val)],
            axis=join_axis,
        )[indices_val]
        index_logps = scipy_logprob(i_val, p_val)
        exp_obs_logps = component_logps + index_logps[(Ellipsis,) + (None,) * join_axis]

        m_val = np.stack([x_val, y_val, z_val], axis=join_axis)[indices_val]

        I_logp_vals = I_logp_fn(p_val, i_val)
        M_logp_vals = M_logp_fn(m_val, i_val)

        logp_vals = M_logp_vals + I_logp_vals[(Ellipsis,) + (None,) * join_axis]

        np.testing.assert_almost_equal(logp_vals, exp_obs_logps, decimal=decimals)


@pytest.mark.parametrize(
    "A_parts, indices",
    [
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (np.array([[0, 1], [2, 2]]), slice(2, 3)),
        ),
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (slice(2, 3), np.array([[0, 1], [2, 2]])),
        ),
        (
            (
                np.random.normal(size=(5, 4, 3)),
                np.random.normal(size=(5, 4, 3)),
                np.random.normal(size=(5, 4, 3)),
            ),
            (
                np.array([[0], [2], [1]]),
                slice(None),
                np.array([2, 1]),
                slice(2, 3),
            ),
        ),
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (slice(2, 3), np.array([0, 1, 2])),
        ),
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (np.array([[0, 1], [2, 2]]), np.array([[0, 1], [2, 2]])),
        ),
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (
                np.array([[0, 1], [2, 2]]),
                np.array([[0, 1], [2, 2]]),
                np.array([[0, 1], [2, 2]]),
            ),
        ),
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (np.array([[0, 1], [2, 2]]), np.array([[0, 1], [2, 2]]), 1),
        ),
        (
            (
                np.random.normal(size=(5, 4, 3)),
                np.random.normal(size=(5, 4, 3)),
            ),
            (slice(0, 2),),
        ),
        (
            (
                np.random.normal(size=(5, 4, 3)),
                np.random.normal(size=(5, 4, 3)),
            ),
            (slice(0, 2), np.random.randint(3, size=(2, 3))),
        ),
    ],
)
def test_expand_indices_basic(A_parts, indices):
    A = at.stack(A_parts)
    at_indices = [as_index_constant(idx) for idx in indices]
    full_indices = expand_indices(at_indices, shape_tuple(A))
    assert len(full_indices) == A.ndim
    exp_res = A[indices].eval()
    res = A[full_indices].eval()
    assert np.array_equal(res, exp_res)


@pytest.mark.parametrize(
    "A_parts, indices",
    [
        (
            (
                np.random.normal(size=(6, 5, 4, 3)),
                np.random.normal(size=(6, 5, 4, 3)),
                np.random.normal(size=(6, 5, 4, 3)),
            ),
            (
                slice(None),
                np.array([[0], [2], [1]]),
                slice(None),
                np.array([2, 1]),
                slice(2, 3),
            ),
        ),
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (np.array([[0, 1], [2, 2]]), slice(None), np.array([[0, 1], [2, 2]])),
        ),
    ],
)
def test_expand_indices_moved_subspaces(A_parts, indices):
    A = at.stack(A_parts)
    at_indices = [as_index_constant(idx) for idx in indices]
    full_indices = expand_indices(at_indices, shape_tuple(A))
    assert len(full_indices) == A.ndim
    exp_res = A[indices].eval()
    res = A[full_indices].eval()
    assert np.array_equal(res, exp_res)


@pytest.mark.parametrize(
    "A_parts, indices",
    [
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (slice(2, 3), np.array([0, 1, 2]), 1),
        ),
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (slice(2, 3), 1, np.array([0, 1, 2])),
        ),
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (1, slice(2, 3), np.array([0, 1, 2])),
        ),
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (np.random.randint(2, size=(4, 3)), 1, 0),
        ),
    ],
)
def test_expand_indices_single_indices(A_parts, indices):
    A = at.stack(A_parts)
    at_indices = [as_index_constant(idx) for idx in indices]
    full_indices = expand_indices(at_indices, shape_tuple(A))
    assert len(full_indices) == A.ndim
    exp_res = A[indices].eval()
    res = A[full_indices].eval()
    assert np.array_equal(res, exp_res)


@pytest.mark.parametrize(
    "A_parts, indices",
    [
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (None,),
        ),
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (None, None, None),
        ),
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (None, 1, None, 0, None),
        ),
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (slice(2, 3), None, 1, None, 0, None),
        ),
        (
            (
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
                np.random.normal(size=(4, 3)),
            ),
            (slice(2, 3), None, 1, 0, None),
        ),
    ],
)
def test_expand_indices_newaxis(A_parts, indices):
    A = at.stack(A_parts)
    at_indices = [as_index_constant(idx) for idx in indices]
    full_indices = expand_indices(at_indices, shape_tuple(A))
    assert len(full_indices) == A.ndim
    exp_res = A[indices].eval()
    res = A[full_indices].eval()
    assert np.array_equal(res, exp_res)


def test_mixture_with_DiracDelta():
    srng = at.random.RandomStream(29833)

    X_rv = srng.normal(0, 1, name="X")
    Y_rv = dirac_delta(0.0)
    Y_rv.name = "Y"

    I_rv = srng.categorical([0.5, 0.5], size=4)

    M_rv = at.stack([X_rv, Y_rv])[I_rv]
    M_rv.name = "M"

    logp_res, _ = conditional_logprob(M_rv, I_rv)

    assert M_rv in logp_res


def test_switch_mixture():
    srng = at.random.RandomStream(29833)

    X_rv = srng.normal(-10.0, 0.1, name="X")
    Y_rv = srng.normal(10.0, 0.1, name="Y")

    I_rv = srng.bernoulli(0.5, name="I")
    i_vv = I_rv.clone()
    i_vv.name = "i"

    Z1_rv = at.switch(I_rv, X_rv, Y_rv)
    z_vv = Z1_rv.clone()
    z_vv.name = "z1"

    fgraph, *_ = construct_ir_fgraph({Z1_rv: z_vv, I_rv: i_vv})

    out_rv = fgraph.outputs[0].owner.inputs[0]
    assert isinstance(out_rv.owner.op, MixtureRV)
    assert not hasattr(
        fgraph.outputs[0].tag, "test_value"
    )  # aesara.config.compute_test_value == "off"
    assert fgraph.outputs[0].name is None

    Z1_rv.name = "Z1"

    fgraph, *_ = construct_ir_fgraph({Z1_rv: z_vv, I_rv: i_vv})

    out_rv = fgraph.outputs[0].owner.inputs[0]
    assert out_rv.name == "Z1-mixture"

    # building the identical graph but with a stack to check that mixture computations are identical

    Z2_rv = at.stack((X_rv, Y_rv))[I_rv]

    fgraph2, *_ = construct_ir_fgraph({Z2_rv: z_vv, I_rv: i_vv})

    assert equal_computations(fgraph.outputs, fgraph2.outputs)

    z1_logp, _ = joint_logprob(realized={Z1_rv: z_vv, I_rv: i_vv})
    z2_logp, _ = joint_logprob(realized={Z2_rv: z_vv, I_rv: i_vv})

    # below should follow immediately from the equal_computations assertion above
    assert equal_computations([z1_logp], [z2_logp])

    np.testing.assert_almost_equal(0.69049938, z1_logp.eval({z_vv: -10, i_vv: 0}))
    np.testing.assert_almost_equal(0.69049938, z2_logp.eval({z_vv: -10, i_vv: 0}))


def test_dirac_delta_mixture_dtype():
    srng = at.random.RandomStream()

    p = at.scalar("p")
    I_rv = srng.bernoulli(p)

    X_rv = dirac_delta(at.as_tensor(0))
    assert X_rv.type.dtype.startswith("int")

    Y_rv = srng.poisson(1.0)

    Z_rv = at.stack([X_rv, Y_rv])[I_rv]
    logprob, (
        i_vv,
        z_vv,
    ) = joint_logprob(I_rv, Z_rv)

    assert logprob.type.dtype.startswith("float")
